// ==========  This file is under  LGPL, the GNU Lesser General Public Licence
// ==========  Dialing Lemmatizer (www.aot.ru)
// ==========  Copyright by Alexey Sokirko
// ==========  Changed by Pavel Makarov (January 2007), e-mail: PavMakarov@gmail.com

#ifndef AhoKorasickMorph_h
#define AhoKorasickMorph_h

#include "MorphAutomat.h"
#include "FormInfo.h"

//       
struct CLemmaInfoAndLemma 
{
	//       morph.bases
	int			m_LemmaStrNo;

	//   :   ,   ,   
	CLemmaInfo	m_LemmaInfo;

	bool operator < (const CLemmaInfoAndLemma& X) const 
	{
		if (m_LemmaInfo.m_FlexiaModelNo != X.m_LemmaInfo.m_FlexiaModelNo)
			return m_LemmaInfo.m_FlexiaModelNo < X.m_LemmaInfo.m_FlexiaModelNo;

		return m_LemmaStrNo < X.m_LemmaStrNo;
	}

};

class CMorphDict   
{
protected:
	
	CMorphAutomat*				m_pFormAutomat;

	// CMorphDict::m_ModelsIndex - ,   CMorphDict::m_LemmaInfos,
	//   m_LemmaInfo.m_FlexiaModelNo.
	// CMorphDict::m_ModelsIndex    CMorphDict::m_ModelsIndex
	//   m_LemmaInfo.m_FlexiaModelNo.  a=m_ModelsIndex[i]  b=m_ModelsIndex[i+1],
	//    j (a<=j<b): LemmaInfos[j].m_LemmaInfo.m_FlexiaModelNo == i.
	// :     ,     ,
	//    .
	vector<int>					m_ModelsIndex;
	
	struct IsLessMorphInterp 
	{
		const CShortStringHolder& m_SearchInfos;

		IsLessMorphInterp(const CShortStringHolder& SearchInfos) : m_SearchInfos(SearchInfos) {}

		bool	operator () (const CLemmaInfoAndLemma& _X1, const char* _X2) const
		{
			const char* base1 = m_SearchInfos[_X1.m_LemmaStrNo].GetString();

			return strcmp(base1, _X2) < 0;
		}
		bool	operator () (const char* _X1, const CLemmaInfoAndLemma& _X2) const
		{
			const char* base2 = m_SearchInfos[_X2.m_LemmaStrNo].GetString();

			return strcmp(_X1, base2) < 0;
		}
		bool	operator () (const CLemmaInfoAndLemma& _X1, const CLemmaInfoAndLemma& _X2) const
		{
			const char* base1 = m_SearchInfos[_X1.m_LemmaStrNo].GetString();
			const char* base2 = m_SearchInfos[_X2.m_LemmaStrNo].GetString();

			return strcmp(base1, base2) < 0;
		}
	};

	IsLessMorphInterp m_SearchInfoLess;

	//       (. CAutomAnnotationInner):
	//   ,    ,  ,
	//      m_LemmaInfos
	void	GetLemmaInfos (const string& Text, size_t TextPos, vector<CAutomAnnotationInner>& Infos) const;

	//   (. m_ModelsIndex)
	void	CreateModelsIndex();

public:
	//  : , , ,  
	vector<CFlexiaModel>		m_FlexiaModels;	
	
	//  :      
	vector<CAccentModel>		m_AccentModels;
	
	// 
	CShortStringHolder			m_Bases;

	//   :       morph.bases ( m_Bases),
	//   :   ,   ,  
	vector<CLemmaInfoAndLemma>	m_LemmaInfos;

	//  
	StringVector				m_Prefixes;

	// ,    .      .
	//       morph.annot  
	vector<BYTE>				m_NPSs;

	CMorphDict(MorphLanguageEnum Language);
	virtual ~CMorphDict();
	void	InitAutomat(CMorphAutomat*	pFormAutomat);

	//   
	bool	Load(wstring wsGrammarFileName);
};

#endif // AhoKorasickMorph_h

