// ==========  This file is under  LGPL, the GNU Lesser General Public Licence
// ==========  Dialing Lemmatizer (www.aot.ru)
// ==========  Changed by Pavel Makarov (January 2007), e-mail: PavMakarov@gmail.com

#include "stdafx.h"

#include "utilit.h"
#include "util_classes.h"

#ifdef DETECT_MEMORY_LEAK
	#ifdef _DEBUG
	#define new DEBUG_NEW
	#undef THIS_FILE
	static char THIS_FILE[] = __FILE__;
	#endif
#endif

const WORD fWordDelim = 1;
const WORD RusUpper = 2;
const WORD RusLower = 4;
const WORD GerUpper = 8;
const WORD GerLower = 16;
const WORD EngUpper = 32;
const WORD EngLower = 64;
const WORD OpnBrck = 128;
const WORD ClsBrck = 256;
const WORD UpRomDigits = 512;
const WORD LwRomDigits = 1024;
const WORD LatinVowel = 2048;
const WORD RussianVowel = 4096;
const WORD URL_CHAR = 8192;

const WORD ASCII[256] = {
/*0   null (nul)*/                                     fWordDelim,
/*1   start of heading (soh)*/                         fWordDelim,
/*2   start of text (stx)*/                            fWordDelim,
/*3   end of text (etx)*/                              fWordDelim,
/*4   end of transmission (eot)*/                      fWordDelim,
/*5   enquiry (enq)*/                                  fWordDelim,
/*6   acknowledge (ack)*/                              fWordDelim,
/*7   bell (bel)*/                                     fWordDelim,
/*8   backspace (bs)*/                                 fWordDelim,
/*9   character tabulation (ht)*/                      fWordDelim,
/*10  line feed (lf)*/                                 fWordDelim,
/*11  line tabulation (vt)*/                           fWordDelim,
/*12  form feed (ff)*/                                 fWordDelim,
/*13  carriage return (cr)*/                           fWordDelim,
/*14  shift out (so)*/                                 fWordDelim,
/*15  shift in (si)*/                                  fWordDelim,
/*16  datalink escape (dle)*/                          fWordDelim,
/*17  device control one (dc1)*/                       fWordDelim,
/*18  device control two (dc2)*/                       fWordDelim,
/*19  device control three (dc3)*/                     fWordDelim,
/*20  device control four (dc4)*/                      fWordDelim,
/*21  negative acknowledge (nak)*/                     fWordDelim,
/*22  syncronous idle (syn)*/                          fWordDelim,
/*23  end of transmission block (etb)*/                fWordDelim,
/*24  cancel (can)*/                                   fWordDelim,
/*25  end of medium (em)*/                             fWordDelim,
/*26  substitute (sub)*/                               fWordDelim,
/*27  escape (esc)*/                                   fWordDelim,
/*28  file separator (is4)*/                           fWordDelim,
/*29  group separator (is3)*/                          fWordDelim,
/*30  record separator (is2)*/                         fWordDelim,
/*31  unit separator (is1)*/                           fWordDelim,
/*32  space ' '*/                                      fWordDelim,
/*33  exclamation mark '!'*/                           fWordDelim|URL_CHAR,
/*34  quotation mark '"'*/                             fWordDelim,
/*35  number sign '#'*/                                fWordDelim,
/*36  dollar sign '$'*/                                fWordDelim|URL_CHAR,
/*37  percent sign '%'*/                               fWordDelim|URL_CHAR,
/*38  ampersand '&'*/                                  fWordDelim|URL_CHAR,
/*39  apostrophe '''*/                                 fWordDelim|RusUpper|RusLower, //     
/*40  left parenthesis '('*/                           fWordDelim|OpnBrck|URL_CHAR,
/*41  right parenthesis ')'*/                          fWordDelim|ClsBrck|URL_CHAR,
/*42  asterisk '*'*/                                   fWordDelim|URL_CHAR,
/*43  plus sign '+'*/                                  fWordDelim|URL_CHAR,
/*44  comma ','*/                                      fWordDelim|URL_CHAR,
/*45  hyphen-minus '-'*/                               fWordDelim|URL_CHAR,
/*46  full stop '.'*/                                  fWordDelim|URL_CHAR,
/*47  solidus '/'*/                                    fWordDelim|URL_CHAR,
/*48  digit zero '0'*/                                 URL_CHAR,
/*49  digit one '1'*/                                  URL_CHAR,
/*50  digit two '2'*/                                  URL_CHAR,
/*51  digit three '3'*/                                URL_CHAR,
/*52  digit four '4'*/                                 URL_CHAR,
/*53  digit five '5'*/                                 URL_CHAR,
/*54  digit six '6'*/                                  URL_CHAR,
/*55  digit seven '7'*/                                URL_CHAR,
/*56  digit eight '8'*/                                URL_CHAR,
/*57  digit nine '9'*/                                 URL_CHAR,
/*58  colon ':'*/                                      fWordDelim|URL_CHAR,
/*59  semicolon ';'*/                                  fWordDelim|URL_CHAR,
/*60  less-than sign '<'*/                             fWordDelim|OpnBrck,
/*61  equals sign '='*/                                fWordDelim|URL_CHAR,
/*62  greater-than sign '>'*/                          fWordDelim|ClsBrck,
/*63  question mark '?'*/                              fWordDelim|URL_CHAR,
/*64  commercial at '@'*/                              fWordDelim|URL_CHAR,
/*65  latin capital letter a 'A'*/                     GerUpper|EngUpper|LatinVowel,
/*66  latin capital letter b 'B'*/                     GerUpper|EngUpper,
/*67  latin capital letter c 'C'*/                     GerUpper|EngUpper,
/*68  latin capital letter d 'D'*/                     GerUpper|EngUpper,
/*69  latin capital letter e 'E'*/                     GerUpper|EngUpper|LatinVowel,
/*70  latin capital letter f 'F'*/                     GerUpper|EngUpper,
/*71  latin capital letter g 'G'*/                     GerUpper|EngUpper,
/*72  latin capital letter h 'H'*/                     GerUpper|EngUpper,
/*73  latin capital letter i 'I'*/                     GerUpper|EngUpper|UpRomDigits|LatinVowel,
/*74  latin capital letter j 'J'*/                     GerUpper|EngUpper,
/*75  latin capital letter k 'K'*/                     GerUpper|EngUpper,
/*76  latin capital letter l 'L'*/                     GerUpper|EngUpper|UpRomDigits,
/*77  latin capital letter m 'M'*/                     GerUpper|EngUpper,
/*78  latin capital letter n 'N'*/                     GerUpper|EngUpper,
/*79  latin capital letter o 'O'*/                     GerUpper|EngUpper|LatinVowel,
/*80  latin capital letter p 'P'*/                     GerUpper|EngUpper,
/*81  latin capital letter q 'Q'*/                     GerUpper|EngUpper,
/*82  latin capital letter r 'R'*/                     GerUpper|EngUpper,
/*83  latin capital letter s 'S'*/                     GerUpper|EngUpper,
/*84  latin capital letter t 'T'*/                     GerUpper|EngUpper,
/*85  latin capital letter u 'U'*/                     GerUpper|EngUpper|LatinVowel,
/*86  latin capital letter v 'V'*/                     GerUpper|EngUpper|UpRomDigits,
/*87  latin capital letter w 'W'*/                     GerUpper|EngUpper,
/*88  latin capital letter x 'X'*/                     GerUpper|EngUpper|UpRomDigits,
/*89  latin capital letter y 'Y'*/                     GerUpper|EngUpper,
/*90  latin capital letter z 'Z'*/                     GerUpper|EngUpper,
/*91  left square bracket '['*/                        fWordDelim|OpnBrck,
/*92  reverse solidus '\'*/                            fWordDelim,
/*93  right square bracket ']'*/                       fWordDelim|ClsBrck,
/*94  circumflex accent '^'*/                          fWordDelim,
/*95  low line '_'*/                                   fWordDelim,
/*96  grave accent '`'*/                               fWordDelim,
/*97  latin small letter a 'a'*/                       GerLower|EngLower|LatinVowel|URL_CHAR,
/*98  latin small letter b 'b'*/                       GerLower|EngLower|URL_CHAR,
/*99  latin small letter c 'c'*/                       GerLower|EngLower|URL_CHAR,
/*100 latin small letter d 'd'*/                       GerLower|EngLower|URL_CHAR,
/*101 latin small letter e 'e'*/                       GerLower|EngLower|LatinVowel|URL_CHAR,
/*102 latin small letter f 'f'*/                       GerLower|EngLower|URL_CHAR,
/*103 latin small letter g 'g'*/                       GerLower|EngLower|URL_CHAR,
/*104 latin small letter h 'h'*/                       GerLower|EngLower|URL_CHAR,
/*105 latin small letter i 'i'*/                       GerLower|EngLower|LwRomDigits|LatinVowel|URL_CHAR,
/*106 latin small letter j 'j'*/                       GerLower|EngLower|URL_CHAR,
/*107 latin small letter k 'k'*/                       GerLower|EngLower|URL_CHAR,
/*108 latin small letter l 'l'*/                       GerLower|EngLower|LwRomDigits|URL_CHAR,
/*109 latin small letter m 'm'*/                       GerLower|EngLower|URL_CHAR,
/*110 latin small letter n 'n'*/                       GerLower|EngLower|URL_CHAR,
/*111 latin small letter o 'o'*/                       GerLower|EngLower|LatinVowel|URL_CHAR,
/*112 latin small letter p 'p'*/                       GerLower|EngLower|URL_CHAR,
/*113 latin small letter q 'q'*/                       GerLower|EngLower|URL_CHAR,
/*114 latin small letter r 'r'*/                       GerLower|EngLower|URL_CHAR,
/*115 latin small letter s 's'*/                       GerLower|EngLower|URL_CHAR,
/*116 latin small letter t 't'*/                       GerLower|EngLower|URL_CHAR,
/*117 latin small letter u 'u'*/                       GerLower|EngLower|LatinVowel|URL_CHAR,
/*118 latin small letter v 'v'*/                       GerLower|EngLower|LwRomDigits|URL_CHAR,
/*119 latin small letter w 'w'*/                       GerLower|EngLower|URL_CHAR,
/*120 latin small letter x 'x'*/                       GerLower|EngLower|LwRomDigits|URL_CHAR,
/*121 latin small letter y 'y'*/                       GerLower|EngLower|URL_CHAR,
/*122 latin small letter z 'z'*/                       GerLower|EngLower|URL_CHAR,
/*123 left curly bracket '{'*/                         fWordDelim|OpnBrck,
/*124 vertical line '|'*/                              fWordDelim,
/*125 right curly bracket '}'*/                        fWordDelim|ClsBrck,
/*126 tilde '~'*/                                      fWordDelim,
/*127 delete ''*/                                     0,
/*128 padding character (pad) '_'*/                    fWordDelim,
/*129 high octet preset (hop) '_'*/                    0,
/*130 break permitted here (bph) '''*/                 0,
/*131 no break here (nbh) '_'*/                        fWordDelim,
/*132 index (ind) '"'*/                                0,
/*133 next line (nel) ':'*/                            fWordDelim,
/*134 start of selected area (ssa) '+'*/               fWordDelim,
/*135 end of selected area (esa) '+'*/                 fWordDelim,
/*136 character tabulation set (hts) '_'*/             fWordDelim,
/*137 character tabulation with justification (htj) '%'*/ fWordDelim,
/*138 line tabulation set (vts) '_'*/                  0,
/*139 partial line forward (pld) '<'*/                 fWordDelim,
/*140 partial line backward (plu) '_'*/                fWordDelim,
/*141 reverse line feed (ri) '_'*/                     fWordDelim,
/*142 single-shift two (ss2) '_'*/                     fWordDelim,
/*143 single-shift three (ss3) '_'*/                   fWordDelim,
/*144 device control string (dcs) '_'*/                fWordDelim,
/*145 private use one (pu1) '''*/                      fWordDelim,
/*146 private use two (pu2) '''*/                      fWordDelim,
/*147 set transmit state (sts) '"'*/                   fWordDelim,
/*148 cancel character (cch) '"'*/                     fWordDelim,
/*149 message waiting (mw) ' '*/                       fWordDelim,
/*150 start of guarded area (spa) '-'*/                fWordDelim,
/*151 end of guarded area (epa) '-'*/                  fWordDelim,
/*152 start of string (sos) '_'*/                      fWordDelim,
/*153 single graphic character introducer (sgci) 'T'*/ fWordDelim,
/*154 single character introducer (sci) '_'*/          fWordDelim,
/*155 control sequence introducer (csi) '>'*/          fWordDelim,
/*156 string terminator (st) '_'*/                     fWordDelim,
/*157 operating system command (osc) '_'*/             fWordDelim,
/*158 privacy message (pm) '_'*/                       fWordDelim,
/*159 application program command (apc) '_'*/          fWordDelim,
/*160 no-break space ''*/                             fWordDelim,
/*161 inverted exclamation mark ''*/                  fWordDelim,
/*162 cent sign ''*/                                  fWordDelim,
/*163 pound sign '_'*/                                 fWordDelim,
/*164 currency sign ''*/                              fWordDelim,
/*165 yen sign '_'*/                                   fWordDelim,
/*166 broken bar ''*/                                 fWordDelim,
/*167 section sign ''*/                               fWordDelim,
/*168 diaeresis ''*/                                  fWordDelim|RusUpper|RussianVowel,
/*169 copyright sign 'c'*/                             fWordDelim,
/*170 feminine ordinal indicator ''*/                 fWordDelim,
/*171 left pointing double angle quotation mark '<'*/  fWordDelim,
/*172 not sign ''*/                                   fWordDelim,
/*173 soft hyphen '-'*/                                fWordDelim,
/*174 registered sign 'R'*/                            fWordDelim,
/*175 macron ''*/                                     fWordDelim,
/*176 degree sign ''*/                                fWordDelim,
/*177 plus-minus sign '+'*/                            fWordDelim,
/*178 superscript two '_'*/                            fWordDelim,
/*179 superscript three '_'*/                          fWordDelim,
/*180 acute '_'*/                                      fWordDelim,
/*181 micro sign ''*/                                 fWordDelim|GerLower|GerUpper,
/*182 pilcrow sign ''*/                               fWordDelim,
/*183 middle dot ''*/                                 fWordDelim,
/*184 cedilla ''*/                                    RusLower|RussianVowel,
/*185 superscript one ''*/                            fWordDelim,
/*186 masculine ordinal indicator ''*/                fWordDelim,
/*187 right pointing double angle quotation mark '>'*/ fWordDelim,
/*188 vulgar fraction one quarter '_'*/                fWordDelim,
/*189 vulgar fraction one half '_'*/                   fWordDelim,
/*190 vulgar fraction three quarters '_'*/             fWordDelim,
/*191 inverted question mark ''*/                     fWordDelim,
/*192 latin capital letter a with grave ''*/          RusUpper|RussianVowel,
/*193 latin capital letter a with acute ''*/          RusUpper,
/*194 latin capital letter a with circumflex ''*/     RusUpper|GerUpper|EngUpper|LatinVowel,
/*195 latin capital letter a with tilde ''*/          RusUpper,
/*196 latin capital letter a with diaeresis ''*/      RusUpper|GerUpper|LatinVowel,
/*197 latin capital letter a with ring above ''*/     RusUpper|RussianVowel,
/*198 latin capital ligature ae ''*/                  RusUpper,
/*199 latin capital letter c with cedilla ''*/        RusUpper|GerUpper|EngUpper,
/*200 latin capital letter e with grave ''*/          RusUpper|GerUpper|EngUpper|LatinVowel|RussianVowel,
/*201 latin capital letter e with acute ''*/          RusUpper|GerUpper|EngUpper|LatinVowel,
/*202 latin capital letter e with circumflex ''*/     RusUpper|GerUpper|EngUpper|LatinVowel,
/*203 latin capital letter e with diaeresis ''*/      RusUpper,
/*204 latin capital letter i with grave ''*/          RusUpper,
/*205 latin capital letter i with acute ''*/          RusUpper,
/*206 latin capital letter i with circumflex ''*/     RusUpper|RussianVowel,
/*207 latin capital letter i with diaeresis ''*/      RusUpper,
/*208 latin capital letter eth (icelandic) ''*/       RusUpper,
/*209 latin capital letter n with tilde ''*/          RusUpper|GerUpper|EngUpper,
/*210 latin capital letter o with grave ''*/          RusUpper,
/*211 latin capital letter o with acute ''*/          RusUpper|RussianVowel,
/*212 latin capital letter o with circumflex ''*/     RusUpper|GerUpper|EngUpper|LatinVowel,
/*213 latin capital letter o with tilde ''*/          RusUpper,
/*214 latin capital letter o with diaeresis ''*/      RusUpper|GerUpper|EngUpper|LatinVowel,
/*215 multiplication sign ''*/                        RusUpper,
/*216 latin capital letter o with stroke ''*/         RusUpper|UpRomDigits,
/*217 latin capital letter u with grave ''*/          RusUpper,
/*218 latin capital letter u with acute ''*/          RusUpper,
/*219 latin capital letter u with circumflex ''*/     RusUpper|GerUpper|EngUpper|LatinVowel|RussianVowel,
/*220 latin capital letter u with diaeresis ''*/      RusUpper|GerUpper|LatinVowel,
/*221 latin capital letter y with acute ''*/          RusUpper|RussianVowel,
/*222 latin capital letter thorn (icelandic) ''*/     RusUpper|RussianVowel,
/*223 latin small letter sharp s (german) ''*/        RusUpper|GerLower|GerUpper|RussianVowel,
/*224 latin small letter a with grave ''*/            RusLower|RussianVowel,
/*225 latin small letter a with acute ''*/            RusLower,
/*226 latin small letter a with circumflex ''*/       RusLower|GerLower|EngLower|LatinVowel,
/*227 latin small letter a with tilde ''*/            RusLower,
/*228 latin small letter a with diaeresis ''*/        RusLower|GerLower|LatinVowel,
/*229 latin small letter a with ring above ''*/       RusLower|RussianVowel,
/*230 latin small ligature ae ''*/                    RusLower,
/*231 latin small letter c with cedilla ''*/          RusLower|GerLower|EngLower,
/*232 latin small letter e with grave ''*/            RusLower|GerLower|EngLower|LatinVowel|RussianVowel,
/*233 latin small letter e with acute ''*/            RusLower|GerLower|EngLower|LatinVowel,
/*234 latin small letter e with circumflex ''*/       RusLower|GerLower|EngLower|LatinVowel,
/*235 latin small letter e with diaeresis ''*/        RusLower,
/*236 latin small letter i with grave ''*/            RusLower,
/*237 latin small letter i with acute ''*/            RusLower,
/*238 latin small letter i with circumflex ''*/       RusLower|RussianVowel,
/*239 latin small letter i with diaeresis ''*/        RusLower,
/*240 latin small letter eth (icelandic) ''*/         RusLower,
/*241 latin small letter n with tilde ''*/            RusLower|GerLower|EngLower,
/*242 latin small letter o with grave ''*/            RusLower,
/*243 latin small letter o with acute ''*/            RusLower|RussianVowel,
/*244 latin small letter o with circumflex ''*/       RusLower|GerLower|EngLower|LatinVowel,
/*245 latin small letter o with tilde ''*/            RusLower,
/*246 latin small letter o with diaeresis ''*/        RusLower|GerLower|EngLower|LatinVowel,
/*247 division sign ''*/                              RusLower,
/*248 latin small letter o with stroke ''*/           RusLower,
/*249 latin small letter u with grave ''*/            RusLower,
/*250 latin small letter u with acute ''*/            RusLower,
/*251 latin small letter u with circumflex ''*/       RusLower|GerLower|EngLower|LatinVowel|RussianVowel,
/*252 latin small letter u with diaeresis ''*/        RusLower|GerLower|LatinVowel,
/*253 latin small letter y with acute ''*/            RusLower|RussianVowel,
/*254 latin small letter thorn (icelandic) ''*/       RusLower|RussianVowel,
/*255 latin small letter y with diaeresis  ''*/       RusLower|RussianVowel
};

file_off_t FileSize (const wchar_t *filename)
{
	FILE * fp = _wfopen (filename, L"rb");
	if (fp == NULL) return 0;
	if (!FSeek(fp, 0,  SEEK_END))
	{
		wstring wsfn(filename);
		string sfn;
		ConvertWStrToStrSysLoc(wsfn, sfn);
		throw CExpc (Format("Cannot move to the begining of the file: %s", sfn.c_str()));;
	}

	file_off_t size = FTell (fp);
	fclose (fp);
	return size;
}

bool FSeek(FILE* fp, file_off_t pos, int origin)
{
	return fseek(fp, pos, origin) == 0;
}

file_off_t FTell(FILE* fp)
{
	return ftell(fp);
}

void(*GlobalErrorMessage)(const string&) = 0;

void ErrorMessage (const string& Titul, const string& Message)
{
	if (GlobalErrorMessage)
	{
		string q = Titul + ":"+Message;
		if (q.empty()|| (q[q.length() -1]!='\n'))	q+='\n';
		GlobalErrorMessage(q);
		return;
	}
	
	#ifndef _SILENT
		#ifndef _CONSOLE
			::MessageBox(0, Message.c_str(), Titul.c_str(), MB_OK);
			return;
		#endif

		string q = Message;
		if (q.empty()|| (q[q.length() -1]!='\n'))	q+='\n';
		fprintf(stderr, "%s: %s", Titul.c_str(), q.c_str());
	#endif
}

void ErrorMessage (const string& Message)
{
	ErrorMessage("error", Message);
}

wstring MakeFName (const wstring& InpitFileName,  const wstring& Ext)
{
	wstring Result = InpitFileName;
	size_t i = Result.find_last_of(L'.'); //   
	if ( i  != string::npos) 
		Result.erase(i);	
	Result += wstring(L".");
	Result += Ext;
	return Result;
}

void rml_TRACE( const char* format, ... )
{
#ifdef _DEBUG
	  va_list arglst;
      char s[2000];
	  if (strlen (format) >  200)
	  {
		  OutputDebugString("!!!!! too long debug line!!!");
		  return;
	  }

      va_start( arglst, format );
	  vsprintf( s, format, arglst);
	  va_end( arglst );
	  OutputDebugString(s);
#endif
}

bool GetLanguageByString (string s, MorphLanguageEnum& Result)
{
	
	EngRusMakeUpper(s);

	if(s == "RUSSIAN"){
		Result = morphRussian;
		return true;
	}
	else if(s == "ENGLISH"){
		Result = morphEnglish;
		return true;
	}
	else if(s == "GERMAN"){
		Result = morphGerman;
		return true;
	}
	else if(s == "GENERIC"){
		Result = morphGeneric;
		return true;
	}
	else if(s == "URL_ABC"){
		Result = morphURL;
		return true;
	}
	else 
		return false;
}

string GetStringByLanguage (MorphLanguageEnum Langua)
{
	switch (Langua) {
		case morphRussian: return  "Russian";
		case morphEnglish: return  "English";
		case morphGerman: return  "German";
		case morphGeneric: return  "Generic";
		case morphURL: return  "URL_ABC";
		default :return "unk";
	}
}

//=================================================================
//===================  STRINGS =====================================
//=================================================================

//====================================================
//============= ENGLISH                   ============
//====================================================

const BYTE  Ccedille = 199;
const BYTE  Egravis = 200;
const BYTE  Ezirkun = 202;
const BYTE  Azirkun = 194;
const BYTE  Ozirkun = 212;
const BYTE  Ntilda = 209;
const BYTE  Eakut = 201;
const BYTE  Uzirkun = 219;

const BYTE  ccedille = 231;
const BYTE  egravis = 232;
const BYTE  ezirkun = 234;
const BYTE  azirkun = 226;
const BYTE  ozirkun = 244;
const BYTE  ntilda = 241;
const BYTE  eakut = 233;
const BYTE  uzirkun = 251;

bool is_generic_upper(BYTE x)     
{
 	return		(ASCII[x]&EngUpper)>0;
} 
 
bool is_generic_lower(BYTE x)     
{
	return		(ASCII[x]&EngLower)>0 || x>=128;
}

bool is_generic_alpha(BYTE x)      
{
	return is_english_alpha(x) || x>=128; 
}

bool is_URL_alpha(BYTE x)      
{
	return (ASCII[x]&URL_CHAR)>0; 
}

bool is_english_upper(BYTE x)     
{
	return		(ASCII[x]&EngUpper)>0;
} 

bool is_english_upper_vowel(BYTE x)
{
	return			((ASCII[x] & EngUpper)>0)
				&&	((ASCII[x] & LatinVowel)>0);
}

bool is_english_lower(BYTE x)     
{
	return		(ASCII[x]&EngLower)>0;
} 

bool is_english_lower_vowel(BYTE x)     
{
	return			( (ASCII[x] & EngLower)>0)
				&&	( (ASCII[x] & LatinVowel)>0);
}

bool is_english_alpha(BYTE x)      
{
	return is_english_lower(x) || is_english_upper(x); 
}

BYTE etoupper (BYTE ch)
{
	if (ch>=(BYTE) 'a' && ch<= (BYTE) 'z')
		return ch -= 'a'-'A';
	else
	switch (ch) {
		case ccedille : return Ccedille;// C with cedille
		case egravis : return Egravis;  // E with gravis
		case ezirkun : return Ezirkun;  // E with zirkun
		case azirkun : return Azirkun;  // A with zirkun
		case ozirkun : return Ozirkun;  // O with zirkun
		case ouml : return Ouml;		// O umlaut
		case ntilda : return Ntilda;	// N with tilda
		case uzirkun : return Uzirkun;  // U with zirkun
		case eakut : return Eakut;		// E acut
		default : return ch;
	}
}

BYTE etolower (BYTE ch)
{
	if (ch>=(BYTE) 'A' && ch<= (BYTE) 'Z')
		return ch += 'a'-'A';
	else
	switch (ch) {
		case Ccedille : return ccedille;// C with cedille
		case Egravis : return egravis;  // E with gravis
		case Ezirkun : return ezirkun;  // E with zirkun
		case Azirkun : return azirkun;  // A with zirkun
		case Ozirkun : return ozirkun;  // O with zirkun
		case Ouml : return ouml;		// O umlaut
		case Ntilda : return ntilda;	// N with tilda
		case Uzirkun : return uzirkun;  // U with zirkun
		case Eakut : return eakut;		// E acut
		default : return ch;
	}
}

//====================================================
//============= GERMAN                  ============
//====================================================
 
bool is_german_upper(BYTE x)     
{
	return		(ASCII[x]&GerUpper) > 0;
} 

bool is_german_lower(BYTE x)     
{
	return		(ASCII[x]&GerLower) > 0;
} 

bool is_german_lower_vowel(BYTE x)     
{
	return			(ASCII[x] & GerLower)
				&&	(ASCII[x] & LatinVowel);
}

bool is_german_upper_vowel(BYTE x)
{
	return			(ASCII[x] & GerUpper)
				&&	(ASCII[x] & LatinVowel);

}

bool is_german_alpha(BYTE x)      
{
	return is_german_lower(x) || is_german_upper(x); 
}

BYTE gtoupper (BYTE ch) 
{ 
	if (is_english_lower (ch)) 
		return etoupper (ch);
	else
		switch (ch )
		 {
			case auml  : return Auml;;
			case uuml  : return Uuml;
			case ouml : return Ouml;
			default : return ch;
		 }
}

BYTE gtolower (BYTE ch) 
{ 
	if (is_english_upper (ch)) 
		return etolower (ch);
	else
		switch (ch )
		 {
			case Auml : return auml;;
			case Uuml  : return uuml;
			case Ouml  : return ouml;
			default : return ch;
		 }
}

//====================================================
//============= RUSSIAN                  ============
//====================================================

bool is_russian_upper(BYTE x)     
{
	return		(ASCII[x]&RusUpper) > 0;
}

bool is_russian_lower(BYTE x)     
{
	return		(ASCII[x]&RusLower) > 0;
}

bool is_russian_lower_vowel(BYTE x)     
{
	return			( (ASCII[x] & RusLower) > 0)
				&&	( (ASCII[x] & RussianVowel) > 0);

}
bool is_russian_upper_vowel(BYTE x)
{
	return			((ASCII[x] & RusUpper) > 0)
				&&	((ASCII[x] & RussianVowel) > 0);
}

bool is_russian_lower_consonant(BYTE x) 
{
	return !is_russian_lower_vowel(x) &&  is_russian_lower(x);
}

bool is_russian_upper_consonant(BYTE x) 
{
	return !is_russian_upper_vowel(x) &&  is_russian_upper(x);
}

bool is_russian_consonant(BYTE x)  
{
	return is_russian_lower_consonant(x) || is_russian_upper_consonant(x);
}

bool is_russian_alpha(BYTE x)      
{
	return is_russian_lower(x) || is_russian_upper(x); 
}

BYTE rtoupper (BYTE ch) 
{ 
	if (ch == LowerJO) return UpperJO;
	if (ch == Apostrophe ) return Apostrophe;

	if ( is_russian_lower(ch) )
		return  ch -= ''-'';
	else
		return ch;
}

BYTE rtolower (BYTE ch) 
{ 
	if (ch == UpperJO) return LowerJO;
	if (ch == Apostrophe ) return Apostrophe;

	if ( is_russian_upper(ch) )
		return ch += ''-'';
	else
		return ch;
}

//========================================= ===========
//============= General encoding functions ============
//====================================================

bool is_alpha (BYTE x)       
{
	return is_russian_alpha(x) || is_german_alpha(x); 
}


bool is_alpha (BYTE x, MorphLanguageEnum Langua)
{
	switch (Langua)
	{
		case morphRussian : return is_russian_alpha(x);
		case morphEnglish : return is_english_alpha(x);
		case morphGerman : return is_german_alpha(x);
		case morphGeneric : return is_generic_alpha(x);
		case morphURL : return is_URL_alpha(x);
	}
	assert (false);
	return false;
}

bool is_lower_alpha(BYTE x, MorphLanguageEnum Langua)
{
	switch (Langua)
	{
		case morphRussian : return is_russian_lower(x);
		case morphEnglish : return is_english_lower(x);
		case morphGerman : return is_german_lower(x);
		case morphGeneric : return is_generic_lower(x);
		case morphURL : return false;
	}
	assert (false);
	return false;
}

bool is_upper_alpha(BYTE x, MorphLanguageEnum Langua)
{
	switch (Langua)
	{
		case morphRussian : return is_russian_upper(x);
		case morphEnglish : return is_english_upper(x);
		case morphGerman : return is_german_upper(x);
		case morphGeneric : return is_generic_upper(x);
		case morphURL : return false;
	}
	assert (false);
	return false;
}

//       
char* RusMakeUpper (char *word)
{
	return RegisterConverter(word, strlen(word), is_russian_lower, rtoupper);
}
//       
char* EngMakeUpper (char *word)
{
	return RegisterConverter(word, strlen(word), is_english_lower, etoupper);
}

string& EngMakeUpper (string& word)
{
	return RegisterConverter(word, word.length(), is_english_lower, etoupper);
}

string& EngMakeLower (string& word)
{
	return RegisterConverter(word, word.length(), is_english_upper, etolower);
}

char* GerMakeUpper (char *word)
{
	return RegisterConverter(word, strlen(word), is_german_lower, gtoupper);
}

string& GerMakeUpper (string& word)
{
	return RegisterConverter(word, word.length(), is_german_lower, gtoupper);
}

//       
char* RusMakeLower (char *word)
{
	return RegisterConverter(word, strlen(word), is_russian_upper, rtolower);
}

//       
string& EngRusMakeLower (string& word)
{
	size_t len = word.length();

	for( size_t i = 0 ; i < len; i++ )
        if ( is_russian_upper((BYTE)word[i]) )
            word[i] = rtolower ((BYTE)word[i]);
		else
			word[i] = etolower ((BYTE) word[i]);

	return word;
}

//       
char* EngRusMakeLower (char* word)
{
	if (word == 0) return 0;
	size_t len =  strlen(word);
	for( size_t i = 0 ; i < len; i++ )
        if ( is_russian_upper((BYTE)word[i]) )
			word[i] = rtolower ((BYTE)word[i]);
		else
			word[i] = etolower ((BYTE) word[i]);

	return word;
}

char* RmlMakeUpper (char *word, MorphLanguageEnum langua)
{
	if (!word) return word;
	return GerEngRusMakeUpperTemplate(word, langua, strlen(word));
}	

string& RmlMakeUpper (string& word, MorphLanguageEnum langua)
{
	GerEngRusMakeUpperTemplate(word, langua, word.length());
	return word;
}	

string& RmlMakeLower (string& word, MorphLanguageEnum langua)
{
	if (word.length() == 0) return word;

	if (langua == morphGerman)
		return RegisterConverter(word, word.length(), is_german_upper, gtolower);
	else
		return EngRusMakeLower(word);
}	

//       
string&  EngRusMakeUpper (string& word)
{
	size_t len = word.length();
	for( size_t i = 0 ; i < len; i++ )
        if ( is_russian_lower((BYTE)word[i]) )
			word[i] = rtoupper ((BYTE)word[i]);
		else
			word[i] = etoupper ((BYTE) word[i]);

	return word;
}

//       
char*  EngRusMakeUpper (char* word)
{
	if (word == 0) return 0;
	size_t len =  strlen(word);

	for( size_t i = 0 ; i < len; i++ )

        if ( is_russian_lower((BYTE)word[i]) )
			word[i] = rtoupper ((BYTE)word[i]);
		else
         if ( is_english_lower((BYTE)word[i]) )
			word[i] = etoupper ((BYTE) word[i]);

	return word;
}

template<class T>
void ConvertJO2JeTemplate(T& src, size_t Length)
{
	for (size_t i = 0; i < Length; i++)
	{
		if ( ((BYTE)src[i]) == LowerJO)
			src[i] = (BYTE)'';
		else
		
			if ( ( (BYTE)src[i]) == UpperJO)
			src[i] = (BYTE)'';
	}

}

void ConvertJO2Je(char* src, size_t Length)
{
	ConvertJO2JeTemplate(src, Length);
}

void ConvertJO2Je(string& src) 
{

	ConvertJO2JeTemplate(src, src.length());
}

void ConvertJO2Je(char* src) 
{
	ConvertJO2JeTemplate(src, strlen(src));
}

#define VSNPRINTF _vsnprintf
string Format( const char* format, ... )
{
	va_list arglst;
	const size_t  SmallBufferSize = 15000;
    char SmallBuffer[SmallBufferSize+1];

    va_start( arglst, format );
	int	OutputLength = VSNPRINTF( SmallBuffer, SmallBufferSize, format, arglst);

	va_end( arglst );

	if (OutputLength == -1)
	{
		va_start( arglst, format );
		va_end( arglst );
	}

	if (		(OutputLength == -1)
			||	(OutputLength < SmallBufferSize)
		)
		return SmallBuffer;

	if (OutputLength > 10000000) // simlply we don't want to ruin the memory (10 Mb is enough)
	{
		assert (false);
		OutputLength = 10000000;
	}

	char * pBuffer = new char[OutputLength+2];

	if (!pBuffer)
		return SmallBuffer;

	va_start( arglst, format );
	VSNPRINTF( pBuffer, OutputLength+1, format, arglst);
	va_end( arglst );
	string Res = pBuffer;

	delete pBuffer;

	return Res;
}
#undef  VSNPRINTF

string&  TrimLeft (string& str)
{
	if (str.size() == 0) return str;
	size_t i = str.find_first_not_of(" \t\n\r");
	str.erase (0, i);
	return str;
}

string&  TrimRight (string& str)
{
	if (str.size() == 0) return str;
	size_t i = str.find_last_not_of(" \t\n\r");
	str.erase (i+1);
	return str;
}

string& Trim (string& str)
{
	TrimLeft(str);
	TrimRight(str);
	return str;
}

// ============     CShortString and  CShortStringHolder =========================

CShortString::CShortString(vector<char>::const_iterator pData)
{
	m_pStringPointer = pData;
}

BYTE CShortString::GetLength() const
{	
	return	(BYTE)m_pStringPointer[0];	
}

vector<char>::const_iterator	CShortString::GetData() const
{	
	return	m_pStringPointer;	
}

const char*	CShortString::GetString() const
{	
	return	&(m_pStringPointer[1]);	
}

bool  IsLessShortString::operator ()(const CShortString& Item1,	const char* Item2) const
{
	return strcmp(Item1.GetString(), Item2) < 0;
}

bool  IsLessShortString::operator ()(const char* Item1, const CShortString& Item2) const
{
	return strcmp(Item1, Item2.GetString() ) < 0;
}

bool  IsLessShortString::operator ()(const CShortString& Item1,	const CShortString& Item2) const
{
	return strcmp(Item1.GetString(), Item2.GetString()) < 0;
}

void CShortStringHolder::ReadShortStringHolder(wstring wsFileName)
{
	clear();
	file_off_t BufferSize = (int)FileSize(wsFileName.c_str()) - sizeof(size_t);

	FILE* fp = _wfopen(wsFileName.c_str(), L"rb");
	if (!fp) return;
	size_t Count;
	fread ((void *)&Count, 1, sizeof(size_t), fp);
	try {
		ReadVectorInner(fp, m_Buffer, BufferSize);
		fclose(fp);
		fp = 0;
	}
	catch (...)
	{
		if (fp) fclose(fp);
		throw;
	}

	reserve(Count);
	int Offset = 0;
	for (DWORD i=0; i < Count; i++)
	{
		CShortString R(m_Buffer.begin()+Offset);
		push_back(R);
		Offset +=   R.GetLength() + 2;
	}
}

template<class T>
bool CShortStringHolder::CreateFromSequence(T begin, T end)
{
	m_Buffer.clear();
	size_t Count = 0;
	for (; begin != end; begin++)
	{
		size_t length = begin->length();
		if (length > 254)
		{
			string s = *begin + " - too long";
			ErrorMessage(s.c_str(), "Short string convertor");
			return false;
		}

		m_Buffer.push_back((BYTE)length);

		// add with terminating null 
		m_Buffer.insert(m_Buffer.end(), begin->c_str(), begin->c_str() + length+1);

		Count++;
	}
	
	size_t Offset = 0;
	clear();

	for (DWORD i=0; i < Count; i++)
	{
		CShortString R(m_Buffer.begin()+Offset);
		push_back(R);
		Offset +=   R.GetLength() + 2;
	}

	return true;
}

bool LoadFileToString(wstring wsFileName, string& Result)
{
	if (_waccess(wsFileName.c_str(), 4) != 0) return false;

	vector<char> V;
	ReadVector(wsFileName, V);
	Result = string (V.begin(), V.end());

	return true;
}

// ======================== CExcp ================
CExpc::CExpc(const string& Cause, int ErrorCode) 
{
    m_strCause = Cause;
	m_ErrorCode = ErrorCode;
}

CExpc::CExpc(const CExpc& from)
{
	*this = from;
}
CExpc& CExpc::operator= (const CExpc& from)
{
	m_strCause = from.m_strCause;
	return *this;
}

// ====================   StringTokenizer =======================

void StringTokenizer::initialize(const char *_text, const char *_delims)
{
	delims = _delims;
	i = -1;
	size_t leng = strlen(_text);
	text_ptr = new char[leng+2];
	text = text_ptr;
	strcpy(text, _text);
	text[leng+1] = 0;
}

StringTokenizer::StringTokenizer(const char *_text, const char *_delims)
{
	initialize(_text, _delims);
}

StringTokenizer::~StringTokenizer()
{
	delete []text_ptr;
}

const char * StringTokenizer::get_rest () const 
{ 
	return text;	
}
const char * StringTokenizer::val() const  
{
	return _val;
}

int  StringTokenizer::count() const  
{
	return i;
}

bool StringTokenizer::has_next() const 
{
	return (*text) != 0;
}

bool StringTokenizer::is_delim(char ch) const
{
		return delims.find_first_of(ch) != string::npos;
}

void StringTokenizer::reinitialize(const char *_text, const char *_delims)
{
	assert (text_ptr);
	delete []text_ptr;
	initialize(_text, _delims);
}

const char * StringTokenizer::operator ()()
{
	++i;
	// skip delims
	while(*text && is_delim(*text))
		text++;

	if(*text == 0)	return 0;

	char *ret_val = text;

	while(*text && !is_delim(*text))
		text++;
	*text++ = 0;
	_val = ret_val;

	if (_val == 0)
		throw CExpc ("Exception in StringTokenizer::operator ()");

	return ret_val;
}
string StringTokenizer::next_token ()
{
	const char* s = operator()();
	if (s) 
		return s;
	else 
		return string();
}

// redifined operators new and delete are used to find memory leaks
#ifdef DETECT_MEMORY_LEAK
#undef new
	bool bStartTraceAllocation = false;
	void* operator new(size_t nSize, LPCSTR lpszFileName, int nLine)
	{
		void *p = new char[nSize];
		if (bStartTraceAllocation)
			rml_TRACE("allocation %X size %i from file %s line %i\n", p, nSize, lpszFileName, nLine);
		return p;
	}

	void  operator delete(void* p, LPCSTR lpszFileName, int nLine)
	{
		delete p;
	}
#endif



