// ==========  This file is under  LGPL, the GNU Lesser General Public Licence
// ==========  Dialing Syntax Analysis (www.aot.ru)
// ==========  Copyright by Alexey Sokirko
// ==========  Changed by Pavel Makarov (January 2007), e-mail: PavMakarov@gmail.com

#ifndef utilit_h
 #define utilit_h

#include	<stdio.h>

#define WIN32_LEAN_AND_MEAN		// Exclude rarely-used stuff from Windows headers

// it is used to find memory leaks
#ifdef DETECT_MEMORY_LEAK
	extern void* operator new(size_t nSize, const char* lpszFileName, int nLine);
	extern void  operator delete(void* p, const char* lpszFileName, int nLine);
	#define DEBUG_NEW new(THIS_FILE, __LINE__)
#endif

#include	"windows.h"

#include	<assert.h>
#include	<string>
#include	<algorithm>
#include	<vector>
#include	"set"
#include	<io.h>
#include	<fcntl.h>


#pragma warning (disable : 4018)
#pragma warning (disable : 4244)
#pragma warning (disable : 4103)
#pragma warning  (disable : 4530)
#pragma warning  (disable : 4251)
#pragma warning  (disable : 4996)
		
#ifdef STLPORT
	using namespace stlport;
#else
	using namespace std;
#endif

typedef unsigned char BYTE;

typedef vector<string> StringVector;

const BYTE   cHyphenChar   = (BYTE) '-';
const BYTE   cPeriodChar   = (BYTE) '.';
const BYTE   cEofChar      = (BYTE) 0x1a;
const BYTE   cPtrChar      = (BYTE) 0x13;
const BYTE   cSpaceChar    = (BYTE) 32;
const BYTE   cBackChar     = (BYTE) '';
const BYTE   cFullChar    = (BYTE) '';
const BYTE   cNumChar      = (BYTE) 'N';
const BYTE   cAmpChar      = (BYTE) '&';
const BYTE   cParagraph    = (BYTE) 21;
const BYTE   cCompanyChar  = (BYTE) 176;
const BYTE   cIonChar      = (BYTE) 183;
const BYTE   cNumberChar   = (BYTE) 0xB9;
const BYTE   cPiChar       = (BYTE) 182;
const BYTE   cEllipseChar  = (BYTE) 133;

// morphology&syntax
typedef enum {morphUnknown = 0, morphRussian = 1, morphEnglish = 2, morphGerman = 3, morphGeneric = 4, morphURL = 5, morphDigits = 6} MorphLanguageEnum;

typedef int file_off_t;

//     
class CExpc
{
 public:
	string	m_strCause;
	int		m_ErrorCode;

	CExpc(const string& Cause, int ErrorCode = -1);
	CExpc(const CExpc& from);
	CExpc& operator= (const CExpc& from);
};

// working with files
extern file_off_t	FileSize (const wchar_t *filename);
extern bool			FSeek(FILE* fp, file_off_t pos, int origin);
extern file_off_t	FTell(FILE* fp);

extern void		ErrorMessage (const string& Titul, const string& Message);
extern void		ErrorMessage (const string& Message);
extern wstring	MakeFName ( const wstring& InpitFileName,  const wstring& Ext);

// working with strings
extern string	Format( const char* format, ... );
extern string&  TrimLeft (string& str);
extern string&  TrimRight (string& str);
extern string&	Trim (string& str);
extern void		rml_TRACE( const char* format, ... );
extern bool		LoadFileToString(wstring wsFileName, string& Result);

extern bool GetLanguageByString (string s, MorphLanguageEnum& Result);
extern string GetStringByLanguage (MorphLanguageEnum Langua);


// ==============   Restoring vectors  ======================

// ============== DWORD =====================
inline size_t get_size_in_bytes (const DWORD& t)
{
	return sizeof(DWORD);
}

inline size_t restore_from_bytes(DWORD& i, const BYTE* buf)
{
	i = *((DWORD*)buf);
	return get_size_in_bytes(i);
}

// ============== int =====================
inline size_t get_size_in_bytes (const int& t)
{
	return sizeof(int);
}

inline size_t restore_from_bytes(int& i, const BYTE* buf)
{
	i = *((int*)buf);
	return get_size_in_bytes(i);
}

// ============== CHAR =====================
inline size_t get_size_in_bytes (const char& t)
{
	return 1;
}

inline size_t restore_from_bytes(char& i, const BYTE* buf)
{
	i = *((char*)buf);
	return get_size_in_bytes(i);
}

// ============== BYTE =====================
inline size_t get_size_in_bytes (const BYTE& t)
{
	return 1;
}

inline size_t restore_from_bytes(BYTE& i, const BYTE* buf)
{
	i = *((BYTE*)buf);
	return get_size_in_bytes(i);
}

// ============== WORD =====================
inline size_t get_size_in_bytes (const WORD& t)
{
	return 2;
}

inline size_t restore_from_bytes(WORD& i, const BYTE* buf)
{
	i = *((WORD*)buf);
	return get_size_in_bytes(i);
}

// ============== DOUBLE =====================
inline size_t get_size_in_bytes (const double& t)
{
	return sizeof(t);
}

inline size_t restore_from_bytes(double& i, const BYTE* buf)
{
	i = *((double*)buf);
	return get_size_in_bytes(i);
}

const int VectorMaxStructSize = 200;

template <class T>
void ReadVectorInner (FILE* fp, vector<T>& V, size_t Count)
{
#ifdef  WIN32
	if (_setmode( _fileno( fp ), _O_BINARY ) != _O_BINARY) 
		throw CExpc(Format("ReadVectorInner: File is not opened in binary mode!"));
#endif

	try
	{
		T dummy;
		size_t size_of_t = get_size_in_bytes(dummy);

		BYTE buffer[VectorMaxStructSize];
		assert (size_of_t < VectorMaxStructSize);

		V.clear();
		V.reserve(Count);
		if (V.capacity() < Count) 
			throw CExpc(Format("cannot allocate %u bytes in ReadVectorInner",size_of_t*Count));

		for (int i = 0; i < Count; i++)
		{
			if (fread ((void*)buffer, size_of_t, 1, fp)!=1)
				throw CExpc(Format("cannot read %i item in ReadVectorInner",i));
			restore_from_bytes (dummy, buffer);
			V.push_back(dummy);
		}
	}
	catch ( length_error  &e ) 
	{
		fprintf (stderr, "ReadVectorInner:length_error exception is caught: %s\n", e.what( ));
		throw;
	}
	catch ( exception &e ) 
	{
		fprintf (stderr, "ReadVectorInner: general std::exception is caught: %s\n", e.what( ));
		throw;
	}
}

template <class T>
inline void ReadVector (const wstring& wsFileName, vector<T>& V)
{
	V.clear();
	file_off_t sz = FileSize(wsFileName.c_str());

	FILE* fp = _wfopen (wsFileName.c_str(), L"rb");
	if (!fp) return;
 	
	T dummy;
	size_t size_of_t = get_size_in_bytes(dummy);
	size_t Count = (size_t)sz/size_of_t;
	try {
        ReadVectorInner(fp, V, Count);
		fclose (fp);
		fp = 0;
	}
	catch (...)
	{
		if (fp) fclose (fp);
		throw;
	}
}

//  =============  Punctuation Letters ======================
extern  bool is_english_upper(BYTE x);
extern  bool is_english_lower(BYTE x);

const BYTE Auml  = 196; // ""
const BYTE auml  = 228; // ""
const BYTE Uuml  = 220; // ""
const BYTE uuml  = 252; // ""
const BYTE Ouml  = 214; // ""
const BYTE ouml  = 246; // ""
const BYTE szlig  = 223; //""
const BYTE Nu  = 181;   // ""
const BYTE agrave  = 224; //""
const BYTE egrave  = 232; //""
const BYTE eacute  = 233; //""

const BYTE LowerJO  = (BYTE)'';
const BYTE UpperJO  = (BYTE)'';
const BYTE Apostrophe  = 39;
 
extern  bool is_german_upper(BYTE x); 
extern  bool is_german_lower(BYTE x);
extern  bool is_russian_upper(BYTE x);     
extern  bool is_russian_lower(BYTE x);

extern  bool is_english_alpha(BYTE x);
extern  bool is_russian_alpha(BYTE x);
extern  bool is_german_alpha(BYTE x);

extern  bool is_alpha (BYTE x);
extern  bool is_alpha (BYTE x, MorphLanguageEnum langua);
extern  bool is_lower_alpha(BYTE x, MorphLanguageEnum langua);
extern  bool is_upper_alpha(BYTE x, MorphLanguageEnum langua);

// ===============  Register ========================================
extern  BYTE etoupper (BYTE ch);
extern  BYTE etolower (BYTE ch);
extern  BYTE rtoupper (BYTE ch); 
extern  BYTE rtolower (BYTE ch); 
extern  BYTE gtoupper (BYTE ch); 
extern  BYTE gtolower (BYTE ch); 
extern char* RusMakeUpper (char *word);
extern char* EngMakeUpper (char *word);
extern string& EngMakeUpper (string& word);
extern string& EngMakeLower (string& word);
extern char* GerMakeUpper (char *word);
extern string& GerMakeUpper (string& word);
extern char* RusMakeLower (char *word);
extern string& EngRusMakeLower (string& word);
extern char* EngRusMakeLower (char* word);
extern char* RmlMakeUpper (char *word, MorphLanguageEnum langua);
extern string& RmlMakeUpper (string& word, MorphLanguageEnum langua);
extern string& RmlMakeLower (string& word, MorphLanguageEnum langua);

//       
extern string&  EngRusMakeUpper (string& word);
//       
extern char*  EngRusMakeUpper (char* word);

extern void ConvertJO2Je(string& src);
extern void ConvertJO2Je(char* src);
extern void ConvertJO2Je(char* src, size_t Length);

template <class T, class Pred, class Conv>
T& RegisterConverter (T& word, size_t Len, Pred P,  Conv C)
{
	for( size_t i = 0 ; i < Len; i++ )
        if ( P( (BYTE)word[i] )  )
			word[i] = C ( (BYTE)word[i] );

    return word;
}

template <class T>
T& GerEngRusMakeUpperTemplate (T& word, MorphLanguageEnum Langua, size_t Len )
{
	if (Len == 0) return word;

	if (Langua == morphGerman)
		return RegisterConverter(word, Len, is_german_lower, gtoupper);
	else
		for( size_t i = 0 ; i < Len; i++ )
			if (is_russian_lower((BYTE)word[i]))
					word[i] = rtoupper ( (BYTE)word[i] );
			else
			if (is_english_lower((BYTE)word[i]))
					word[i] = etoupper ( (BYTE)word[i] );

	return word;
}

// ===============  Unicode ========================================

//  string  wstring
inline void ConvertStrToWStr(string& str, wstring& wstr)
{
	wchar_t* pwcText = new wchar_t[str.length() + 1];	//      Unicode

	mbstowcs(pwcText, str.c_str(), str.length() + 1);	//  ,   0
	wstr.assign(pwcText);

	delete [] pwcText;
}

//  wstring  string
inline void ConvertWStrToStr(wstring& wstr, string& str)
{
	int nStrLen = wcstombs(NULL, wstr.c_str(), 0) + 1;	//    Multibyte,    0

	char* pcText = new char[nStrLen];					//      Multibyte

	wcstombs(pcText, wstr.c_str(), nStrLen);			//  ,   0			
	str.assign(pcText);

	delete [] pcText;
}

//  string  wstring    
inline void ConvertStrToWStrSysLoc(string& str, wstring& wstr)
{
	string sCurLoc = setlocale(LC_CTYPE, NULL);			//   
	setlocale(LC_CTYPE, "");							//    

	ConvertStrToWStr(str, wstr);						//  string  wstring

	setlocale(LC_CTYPE, sCurLoc.c_str());				//   
}

//  wstring  string    
inline void ConvertWStrToStrSysLoc(wstring& wstr, string& str)
{
	string sCurLoc = setlocale(LC_CTYPE, NULL);			//   
	setlocale(LC_CTYPE, "");							//    

	ConvertWStrToStr(wstr, str);						//  wstring  string

	setlocale(LC_CTYPE, sCurLoc.c_str());				//   
}

#endif	// utilit_h



